from __future__ import annotations

import importlib
import inspect
import posixpath
from collections import deque
from collections.abc import Callable
from types import ModuleType
from typing import Any

import yaml

INSTANTIATE_PREFIX = '@'
RELATIVE_PATH_PREFIX = '.'


class ConfigError(Exception):
    pass


def _to_dict(obj):
    if isinstance(obj, Config):
        return obj._to_dict()
    elif isinstance(obj, dict):
        return {k: _to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_to_dict(v) for v in obj]
    else:
        return obj


def _determine_module_by_path(path: str) -> tuple[str, str]:
    module_path = path.split('.')
    object_path = deque([])

    while len(module_path) > 0:
        try:
            possible_module_path = '.'.join(module_path)
            importlib.import_module(possible_module_path)
            return possible_module_path, '.'.join(object_path)
        except ModuleNotFoundError:
            object_path.appendleft(module_path.pop())

    raise ImportError(f'Module not found for path: {path}')


def _get_object_from_path(module: Any, object_path: str) -> Any:
    x = module
    if object_path:
        for part in object_path.split('.'):
            x = getattr(x, part)
    return x


def _import_object_from_path(path: str) -> Any:
    """
    Import an object from a string path starting with '@'.

    Args:
        path (str): Path to the object in the format "@module.submodule.object"

    Returns:
        The imported object

    Raises:
        ImportError: If the module or object cannot be imported
    """
    assert path.startswith(INSTANTIATE_PREFIX), f"Path must start with '{INSTANTIATE_PREFIX}'"

    # Remove the leading '@'
    path = path[len(INSTANTIATE_PREFIX) :]

    module_path, object_path = _determine_module_by_path(path)

    # Import the module
    module = importlib.import_module(module_path)
    obj = _get_object_from_path(module, object_path)
    return obj


def _get_base_path_from_default(default: Any) -> str:
    """Extract base path from different types of default values."""
    if isinstance(default, Config):
        assert default._creator_module is not None, (
            'Config was created in an unknown module. Probably in IPython interactive shell. '
            'Consider moving the config to a module.'
        )
        module = default._creator_module
        name = getattr(module.__spec__, 'name', None) if hasattr(module, '__spec__') else None
        name = name or module.__name__
        return name + '.' + 'stub_name'
    elif isinstance(default, str):
        return default.lstrip(INSTANTIATE_PREFIX)
    elif hasattr(default, '__module__') and hasattr(default, '__name__'):
        return f'{default.__module__}.{default.__name__}'
    elif hasattr(default, '__class__') and hasattr(default, 'name'):  # Handle Enum values
        enum_class = default.__class__
        return f'{enum_class.__module__}.{enum_class.__qualname__}.{default.name}'
    else:
        raise ValueError(
            'Default value must be Config, import string, an object with __module__ and __name__, or an Enum value'
        )


def _construct_relative_path(value: str, base_path: str):
    leading = 0
    for leading in range(len(value)):
        if value[leading] != RELATIVE_PATH_PREFIX:
            break

    assert '🤷‍♂️' not in value, 'Configuronic does not support relative imports with 🤷‍♂️. Sorry 🤷‍♂️'
    value = '🤷‍♂️' * leading + value[leading:]

    path = base_path + value
    unix_like_path = path.replace('.', '/').replace('🤷‍♂️', '/../')
    unix_like_norm_path = posixpath.normpath(unix_like_path)
    module_path = unix_like_norm_path.replace('/', '.')
    return module_path


def _resolve_relative_import(value: str, default: Any) -> Any:
    """Resolve a relative import path (starting with '.')."""
    if default is None:
        raise ValueError('Relative import used with no default value')

    base_path = _get_base_path_from_default(default)
    new_path = _construct_relative_path(value, base_path)

    try:
        return _import_object_from_path(f'{INSTANTIATE_PREFIX}{new_path}')
    except (ImportError, ModuleNotFoundError, AttributeError) as e:
        # Check if this looks like a filesystem path rather than a module path
        hint = (
            f'If you meant dot-prefixed string "{value}" instead of relative import "{new_path}",'
            'please use indexed override "--key.idx={value}"',
        )
        raise ImportError(f'Failed to resolve relative import {value} to module {new_path}{hint}') from e


def _can_resolve_relative(default: Any | None) -> bool:
    """Return True if `default` provides a base for relative resolution.

    We allow relative resolution when default is:
    - a Config instance (nested config), or
    - an importable object (has __module__ and __name__), or
    - an Enum value (has class and name), or
    - a string that starts with '@' (import path string)
    """
    if isinstance(default, Config):
        return True
    if isinstance(default, str):
        return default.startswith(INSTANTIATE_PREFIX)
    if hasattr(default, '__module__') and hasattr(default, '__name__'):
        return True
    if hasattr(default, '__class__') and hasattr(default, 'name'):
        return True
    return False


def _resolve_value(value: Any, default: Any | None = None, config: Config | None = None) -> Any:
    """Resolve special strings to actual Python objects.

    Supports two prefixes:

    - ``@`` - absolute import path of the object to instantiate
    - ``.`` - relative path resolution when there's a suitable base: a nested
      :class:`Config`, an importable object (class/function), an Enum value, or
      a string starting with ``@``. Otherwise, treat leading-dot strings as
      literals (e.g., ``../data``, ``./file``, ``.env``).

    For lists and dicts:
    - Both absolute (``@``) and relative (``.``) references are resolved recursively at all nesting levels
    - Nested collections inherit the same resolution context from their parent
    """
    if isinstance(value, str):
        if value.startswith(INSTANTIATE_PREFIX):
            if value[len(INSTANTIATE_PREFIX) :].startswith(INSTANTIATE_PREFIX):
                return value[len(INSTANTIATE_PREFIX) :]
            else:
                return _import_object_from_path(value)
        # Only resolve relative imports when `default` provides a valid base.
        # Treat all other leading-dot strings as literals (e.g., '../data', '.env').
        elif value.startswith(RELATIVE_PATH_PREFIX) and _can_resolve_relative(default):
            return _resolve_relative_import(value, default)
        else:
            return value
    elif isinstance(value, list | tuple):
        return type(value)(_resolve_value(item, default=config, config=config) for item in value)
    elif isinstance(value, dict):
        return {k: _resolve_value(v, default=config, config=config) for k, v in value.items()}
    else:
        return value


def _get_value(obj, key):
    if isinstance(obj, Config):
        return obj._get_value(key)
    elif isinstance(obj, list):
        return obj[int(key)]
    elif isinstance(obj, tuple):
        return obj[int(key)]
    elif isinstance(obj, dict):
        return obj[key]
    else:
        raise ConfigError(f'Cannot get value of {obj} with key {key}')


def _set_value(obj, key, value):
    if isinstance(obj, Config):
        obj._set_value(key, value)
    elif isinstance(obj, list):
        index = int(key)
        default = obj[index] if 0 <= index < len(obj) else None
        obj[index] = _resolve_value(value, default)
    elif isinstance(obj, tuple):
        raise NotImplementedError('Overriding tuple values is not implemented')
    elif isinstance(obj, dict):
        default = obj.get(key) if isinstance(obj, dict) else None
        obj[key] = _resolve_value(value, default)
    else:
        raise ConfigError(f'Cannot set value of {obj} with key {key}')


def _get_creator_module() -> ModuleType | None:
    current_frame = inspect.currentframe()
    # current frame: this function
    # current frame back: place where this function is called from
    # current frame back back: place one level upperer
    assert current_frame is not None, 'Current frame is None. Do your python interpreter support frames?'
    assert current_frame.f_back is not None, 'Current frame back is None. Should not happen.'
    assert current_frame.f_back.f_back is not None, (
        'Current frame back back is None. This function was probably called from python interpreter.'
    )

    module = inspect.getmodule(current_frame.f_back.f_back)

    return module


class Config:
    def __init__(self, target, *args, **kwargs):
        """
        Initialize a Config object.

        Stores the callable target and its arguments and keyword arguments, which
        can be overridden/instantiated later.

        The args and kwargs could be strings with special syntax, which will be resolved to actual Python objects.
        "@path.to.object" will be resolved to the object similar to "from path.to import object".

        Relative imports (strings starting with '.') are only resolved during overrides when the
        default value is another Config instance. In all other contexts (including __init__),
        leading-dot strings are treated literally (e.g., '../data', './file', '.env').

        Args:
            target: The target object to be configured.
            *args: Positional arguments to be passed to the target object.
            **kwargs: Keyword arguments to be passed to the target object.

        Raises:
            AssertionError: If the target is not callable.

        Example:
            >>> @cfn.config()
            >>> def sum(a, b):
            >>>     return a + b
            >>> res = sum.override(a=1, b=2).instantiate()
            >>> assert res == 3

            >>> def sum(a, b):
            >>>     return a + b
            >>> res = cfn.Config(sum, a=1, b=2).instantiate()
            >>> assert res == 3

            >>> @cfn.config(status="@http.HTTPStatus.OK")
            >>> def return_status(status):
            >>>     return status
            >>> assert return_status() == http.HTTPStatus.OK
        """
        assert callable(target), f'Target must be callable, got object of type {type(target)}.'
        self.target = target
        self.args = [_resolve_value(arg) for arg in args]  # TODO: cover argument override with tests
        self.kwargs = {}
        self._override_inplace(**kwargs)

        self._creator_module = _get_creator_module()

    def override(self, **overrides) -> Config:
        """
        Create a new Config with updated parameters.

        Supports nested parameter updates using dot notation (e.g. "model.layers").
        Handles absolute imports (@) and relative imports (.).

        Args:
            **overrides: Parameter paths and their new values.

        Returns:
            Config: A new Config with overridden parameters.

        Raises:
            ConfigError: If parameter path is invalid.

        Example:
            >>> cfg = Config(Pipeline, model=Config(MyModel, layers=6))
            >>> new_cfg = cfg.override(**{'model.layers': 12})
            >>> new_cfg = cfg.override(model='@my_models.CustomModel')
        """
        overriden_cfg = self._copy()
        overriden_cfg._override_inplace(**overrides)
        # we want to keep creator module (module override was called from) for the overriden config
        # But we override it after the overrides are applied, so that lists and dicts arguments
        # are resolved relative to the original config, not the overriden config.
        overriden_cfg._creator_module = _get_creator_module()

        return overriden_cfg

    def _override_inplace(self, **overrides):
        for key, value in overrides.items():
            try:
                key_list = key.split('.')

                current_obj = self

                for i, key in enumerate(key_list[:-1]):
                    current_obj = _get_value(current_obj, key)
                    if current_obj is None:
                        path_to_not_found_arg = '.'.join(key_list[: i + 1])
                        raise ConfigError(f"Argument '{path_to_not_found_arg}' not found in config")

                _set_value(current_obj, key_list[-1], value)
            except Exception as e:
                raise ConfigError(f"Failed to override '{key}' with value '{value}'") from e

    def _set_value(self, key, value):
        default = self._get_value(key) if self._has_value(key) else None
        value = _resolve_value(value, default, config=self)

        if key[0].isdigit():
            self.args[int(key)] = value
        else:
            self.kwargs[key] = value

    def _get_value(self, key):
        if key[0].isdigit():
            return self.args[int(key)]
        else:
            return self.kwargs.get(key)

    def _has_value(self, key):
        if key[0].isdigit():
            return int(key) < len(self.args)
        else:
            return key in self.kwargs

    def instantiate(self) -> Any:
        """
        Instatiate the target function with the given arguments and keyword arguments.

        Returns:
            The instantiated target function.

        Raises:
            ConfigError: If the target function cannot be instantiated.
        """
        return self._instantiate_internal()

    def _instantiate_internal(self, path: str = ''):
        """
        Instatiate the target function with the given arguments and keyword arguments.

        Args:
            path (str): The path to the current key. Used for error reporting.

        Returns:
            The instantiated target function.

        Raises:
        """

        def _instantiate_value(value, key, path):
            try:
                if isinstance(value, Config):
                    return value._instantiate_internal(path + f'{key}.')
                elif isinstance(value, list | tuple):
                    return type(value)(_instantiate_value(item, f'{key}[{i}]', path) for i, item in enumerate(value))
                elif isinstance(value, dict):
                    return {k: _instantiate_value(v, f'{key}["{k}"]', path) for k, v in value.items()}
                else:
                    return value
            except Exception as e:
                if isinstance(e, ConfigError):
                    raise e
                else:
                    raise ConfigError(f'Error instantiating "{path}{key}": {e}') from e

        # Recursively instantiate any Config objects in args
        instantiated_args = [_instantiate_value(arg, key, path) for key, arg in enumerate(self.args)]

        # Recursively instantiate any Config objects in kwargs
        instantiated_kwargs = {key: _instantiate_value(value, key, path) for key, value in self.kwargs.items()}

        return self.target(*instantiated_args, **instantiated_kwargs)

    def _to_dict(self) -> dict[str, Any]:
        res = {}

        res['@target'] = f'{INSTANTIATE_PREFIX}{self.target.__module__}.{self.target.__name__}'
        args = [_to_dict(arg) for arg in self.args]
        if len(args) > 0:
            res['*args'] = args
        kwargs = {key: _to_dict(value) for key, value in self.kwargs.items()}
        if len(kwargs) > 0:
            res.update(kwargs)
        return res

    def __str__(self):
        return yaml.dump(self._to_dict(), default_flow_style=None, sort_keys=False, width=140)

    def copy(self) -> Config:
        """
        Recursively copy config signatures.
        """

        cfg = self._copy()
        cfg._creator_module = _get_creator_module()
        return cfg

    def _copy(self):
        """
        Recursively copy config signatures.
        """

        new_args = [arg._copy() if isinstance(arg, Config) else arg for arg in self.args]

        new_kwargs = {key: value._copy() if isinstance(value, Config) else value for key, value in self.kwargs.items()}

        cfg = Config(self.target)
        cfg.args = new_args
        cfg.kwargs = new_kwargs
        cfg._creator_module = self._creator_module
        return cfg

    def __call__(self, **kwargs):
        """
        Override the config with the given kwargs and instantiate the config.

        Useful for creating a function for a CLI.

        Args:
            **kwargs: Keyword arguments to override the config.

        Returns:
            The instantiated config.

        Example:
            >>> import fire
            >>> @cfn.config()
            >>> def sum(a, b):
            >>>     return a + b
            >>> option1 = sum.override(a=1)
            >>> option2 = sum.override(b=2)
            >>> fire.Fire()
            >>> # Shell call: python script.py option1 --b 5
            >>> # Shell call: python script.py option2 --a 5
        """
        return self.override(**kwargs).instantiate()


def config(**kwargs) -> Callable[[Callable], Config]:
    """
    Decorator to create a Config object.

    Args:
        **kwargs: Keyword arguments to be passed to the target object.

    Returns:
        A decorator factory that creates a Config object.

    Example:
        >>> @cfn.config(a=1, b=2)
        >>> def sum(a, b):
        >>>     return a + b
        >>> res = sum.instantiate()
        >>> assert res == 3

        >>> @cfn.config()
        >>> def sum(a, b):
        >>>     return a + b
        >>> res = sum.override(a=1, b=2).instantiate()
        >>> assert res == 3
    """

    def _config_decorator(target):
        config = Config(target, **kwargs)
        config._creator_module = _get_creator_module()
        return config

    return _config_decorator



import subprocess
import sys
from enum import Enum

import pytest

import configuronic as cfn
import tests.support_package.cfg2


class ResolutionEnum(Enum):
    RES_1080P = '1080p'
    RES_720P = '720p'
    RES_480P = '480p'


class Env:
    def __init__(self, camera):
        self.camera = camera


class Camera:
    class InnerResolutionEnum(Enum):
        RES_1080P = '1080p'
        RES_720P = '720p'
        RES_480P = '480p'

    def __init__(self, name: str):
        self.name = name


class MultiEnv:
    def __init__(self, env1: Env, env2: Env):
        self.env1 = env1
        self.env2 = env2


def add(a, b):
    return a + b


def apply(func, a, b):
    return func(a, b)


static_object = Camera(name='Static Camera')


def test_instantiate_class_object_basic_created():
    camera_cfg = cfn.Config(Camera, name='OpenCV')

    camera_obj = camera_cfg.instantiate()

    assert isinstance(camera_obj, Camera)
    assert camera_obj.name == 'OpenCV'


def test_instantiate_class_object_with_function_created():
    add_cfg = cfn.Config(add, a=1, b=2)

    add_obj = add_cfg.instantiate()

    assert add_obj == 3


def test_instantiate_class_object_nested_created():
    camera_cfg = cfn.Config(Camera, name='OpenCV')
    env_cfg = cfn.Config(Env, camera=camera_cfg)

    env_obj = env_cfg.instantiate()

    assert isinstance(env_obj, Env)
    assert isinstance(env_obj.camera, Camera)
    assert env_obj.camera.name == 'OpenCV'


def test_instantiate_class_nested_object_overriden_with_config_created():
    opencv_camera_cfg = cfn.Config(Camera, name='OpenCV')
    luxonis_camera_cfg = cfn.Config(Camera, name='Luxonis')

    env_cfg = cfn.Config(Env, camera=opencv_camera_cfg)

    env_obj = env_cfg.override(camera=luxonis_camera_cfg).instantiate()

    assert isinstance(env_obj, Env)
    assert isinstance(env_obj.camera, Camera)
    assert env_obj.camera.name == 'Luxonis'


def test_instantiate_class_required_args_provided_with_kwargs_override_created():
    incomplete_camera_cfg = cfn.Config(Camera)

    camera_obj = incomplete_camera_cfg.override(name='OpenCV').instantiate()

    assert isinstance(camera_obj, Camera)
    assert camera_obj.name == 'OpenCV'


def test_instantiate_class_required_args_provided_with_path_to_class_created():
    incomplete_env_cfg = cfn.Config(Env)

    env_obj = incomplete_env_cfg.override(camera='@tests.test_config.static_object').instantiate()

    assert isinstance(env_obj, Env)
    assert isinstance(env_obj.camera, Camera)
    assert env_obj.camera.name == 'Static Camera'


def test_instantiate_set_leaf_value_level2_created():
    luxonis_camera_cfg = cfn.Config(Camera, name='Luxonis')
    env1_cfg = cfn.Config(Env, camera=luxonis_camera_cfg)

    env2_cfg = cfn.Config(Env)

    multi_env_cfg = cfn.Config(MultiEnv, env1=env1_cfg, env2=env2_cfg)

    new_camera_cfg = cfn.Config(Camera, name='New Camera')

    full_cfg = multi_env_cfg.override(env2=cfn.Config(Env, camera=new_camera_cfg))
    env_obj = full_cfg.instantiate()

    assert isinstance(env_obj, MultiEnv)
    assert isinstance(env_obj.env1, Env)
    assert isinstance(env_obj.env1.camera, Camera)
    assert env_obj.env1.camera.name == 'Luxonis'
    assert isinstance(env_obj.env2, Env)
    assert isinstance(env_obj.env2.camera, Camera)
    assert env_obj.env2.camera.name == 'New Camera'


def test_override_basic_keeps_original_config():
    cfg = cfn.Config(Camera, name='OpenCV')

    cfg.override(name='New Camera')

    assert cfg.kwargs['name'] == 'OpenCV'


def test_override_nested_keeps_original_config():
    cfg = cfn.Config(
        MultiEnv,
        env1=cfn.Config(Env, camera=cfn.Config(Camera, name='OpenCV')),
        env2=cfn.Config(Env, camera=cfn.Config(Camera, name='Luxonis')),
    )

    cfg.override(env2=cfn.Config(Env, camera=cfn.Config(Camera, name='New Camera')))

    assert cfg.kwargs['env2'].kwargs['camera'].kwargs['name'] == 'Luxonis'


def test_config_non_callable_target_raises_error():
    # TODO: Another posibility is to return the original object in this case
    non_callable = object()

    with pytest.raises(AssertionError):
        cfn.Config(non_callable)


def test_config_to_dict_kwargs_only_produces_correct_dict():
    cfg = cfn.Config(Camera, name='OpenCV')

    expected = {'@target': f'@{Camera.__module__}.{Camera.__name__}', 'name': 'OpenCV'}
    assert cfg._to_dict() == expected


def test_config_to_dict_kwargs_and_args_produces_correct_dict():
    cfg = cfn.Config(add, 1, b=2)

    expected = {'@target': f'@{add.__module__}.{add.__name__}', '*args': [1], 'b': 2}
    assert cfg._to_dict() == expected


def test_config_to_dict_nested_produces_correct_dict():
    cfg = cfn.Config(
        MultiEnv,
        env1=cfn.Config(Env, camera=cfn.Config(Camera, name='OpenCV')),
        env2=cfn.Config(Env, camera=cfn.Config(Camera, name='Luxonis')),
    )

    expected_dict = {
        '@target': f'@{MultiEnv.__module__}.{MultiEnv.__name__}',
        'env1': {
            '@target': f'@{Env.__module__}.{Env.__name__}',
            'camera': {'@target': f'@{Camera.__module__}.{Camera.__name__}', 'name': 'OpenCV'},
        },
        'env2': {
            '@target': f'@{Env.__module__}.{Env.__name__}',
            'camera': {'@target': f'@{Camera.__module__}.{Camera.__name__}', 'name': 'Luxonis'},
        },
    }

    assert cfg._to_dict() == expected_dict


def test_config_str_nested_produces_correct_str():
    cfg = cfn.Config(apply, func=cfn.Config(add, 1, b=2), a=3, b=4)

    # The exact format that matches the actual output from Config.__str__
    expected_str = f"""'@target': '@{apply.__module__}.{apply.__name__}'
func:
  '@target': '@{add.__module__}.{add.__name__}'
  '*args': [1]
  b: 2
a: 3
b: 4
"""

    assert str(cfg) == expected_str


def test_instantiate_not_complete_config_raises_error():
    cfg = cfn.Config(Camera)

    with pytest.raises(TypeError, match="missing 1 required positional argument: 'name'"):
        cfg.instantiate()


def test_config_as_decorator_acts_as_config_class():
    @cfn.config()
    def sum(a, b):
        return a + b

    assert sum.override(a=1, b=2).instantiate() == 3


def test_config_as_decorator_default_args_are_passed_to_target():
    @cfn.config()
    def sum(a=1, b=2):
        return a + b

    assert sum.instantiate() == 3


def test_config_as_decorator_override_values_and_instantiate_works():
    @cfn.config(a=1, b=2)
    def sum(a, b):
        return a + b

    assert sum() == 3


def test_config_callable_works_with_flat_configs():
    @cfn.config(a=1, b=2)
    def sum(a, b):
        return a + b

    assert sum() == 3


def test_instantiate_with_lists_and_dicts():
    # Define some simple callables for testing
    def add(a, b):
        return a + b

    def multiply(a, b):
        return a * b

    def process_list(items):
        return sum(items)

    def process_dict(data):
        return sum(data.values())

    # Create nested configs in args
    nested_in_list = cfn.Config(
        process_list,
        [
            cfn.Config(add, 1, 2),  # This will be 3
            cfn.Config(multiply, 2, 3),  # This will be 6
            5,
        ],
    )

    # Create nested configs in kwargs
    nested_in_dict = cfn.Config(
        process_dict,
        {
            'x': cfn.Config(add, 1, 2),  # This will be 3
            'y': cfn.Config(multiply, 2, 3),  # This will be 6
            'z': 5,
        },
    )

    # Create multi-level nesting
    deeply_nested = cfn.Config(
        add,
        cfn.Config(multiply, 2, 3),  # This will be 6
        cfn.Config(
            process_dict,
            {
                'a': cfn.Config(add, 1, 2),  # This will be 3
                'b': 4,
            },
        ),  # This will be 7
    )  # Final result should be 13

    # Test nested configs in lists
    assert nested_in_list.instantiate() == 14  # 3 + 6 + 5 = 14

    # Test nested configs in dictionaries
    assert nested_in_dict.instantiate() == 14  # 3 + 6 + 5 = 14

    # Test deeply nested configs
    assert deeply_nested.instantiate() == 13  # 6 + 7 = 13

    # Test with decorator syntax and nesting
    @cfn.config()
    def complex_operation(a, items, data):
        return a + sum(items) + sum(data.values())

    result = complex_operation.override(
        a=1, items=[cfn.Config(add, 1, 2), cfn.Config(multiply, 2, 3)], data={'x': cfn.Config(add, 1, 2), 'y': 5}
    ).instantiate()

    assert result == 18  # 1 + (3 + 6) + (3 + 5) = 18


def test_instantiate_exception_during_instantiation_has_correct_path():
    def func(a, b):
        return a + b

    def bad_function(a, b):
        raise ValueError('Bad function')

    cfg = cfn.Config(func, a=cfn.Config(bad_function, a=1, b=2), b=3)

    with pytest.raises(cfn.ConfigError) as e:
        cfg.instantiate()

    assert str(e.value) == 'Error instantiating "a": Bad function'


def test_instantiate_exception_during_instantiation_has_correct_path_with_nested_configs():
    def lvl1_function(lvl1arg):
        return lvl1arg

    def lvl2_function(lvl2arg):
        return lvl2arg

    def bad_function(a, b):
        raise ValueError('Bad function')

    cfg = cfn.Config(
        lvl1_function,
        lvl1_arg=cfn.Config(
            lvl2_function,
            lvl2_arg=cfn.Config(bad_function, a=1, b=2),
        ),
    )

    with pytest.raises(cfn.ConfigError) as e:
        cfg.instantiate()

    assert str(e.value) == 'Error instantiating "lvl1_arg.lvl2_arg": Bad function'


def test_instantiate_exception_during_instantiation_has_correct_path_with_list():
    def func(list_arg):
        return list_arg[0]

    @cfn.config()
    def goob_obj():
        return 1

    @cfn.config()
    def bad_obj():
        raise ValueError('Bad object')

    cfg = cfn.Config(func, list_arg=[goob_obj, bad_obj])

    with pytest.raises(cfn.ConfigError) as e:
        cfg.instantiate()

    assert str(e.value) == 'Error instantiating "list_arg[1]": Bad object'


def test_instantiate_exception_during_instantiation_has_correct_path_with_dict():
    def func(dict_arg):
        return dict_arg['key']

    @cfn.config()
    def goob_obj():
        return 1

    @cfn.config()
    def bad_obj():
        raise ValueError('Bad object')

    cfg = cfn.Config(func, dict_arg={'key': goob_obj, 'bad_key': bad_obj})

    with pytest.raises(cfn.ConfigError) as e:
        cfg.instantiate()

    assert str(e.value) == 'Error instantiating "dict_arg["bad_key"]": Bad object'


def test_instantiate_override_with_complex_path_to_object_works():
    @cfn.config()
    def true_if_math_module(obj):
        import math

        return obj is math

    assert true_if_math_module.override(obj='@math').instantiate()


def test_instantiate_override_with_path_to_module_works():
    @cfn.config()
    def return_true(obj):
        return True

    assert return_true.override(obj='@http.HTTPStatus.OK').instantiate()


def test_override_with_single_dot_relative_path():
    camera_cfg = cfn.Config(Camera, name='OpenCV')
    env_cfg = cfn.Config(Env, camera=camera_cfg)

    env_obj = env_cfg.override(camera='.static_object').instantiate()

    assert env_obj.camera is static_object


def test_override_with_single_dot_enum_inside_class_relative_path():
    @cfn.config(status=Camera.InnerResolutionEnum.RES_1080P)
    def return_value(status):
        return status

    assert return_value.override(status='.RES_720P').instantiate() == Camera.InnerResolutionEnum.RES_720P


def test_override_with_single_dot_enum_relative_path():
    @cfn.config(status=ResolutionEnum.RES_1080P)
    def return_value(status):
        return status

    assert return_value.override(status='.RES_720P').instantiate() == ResolutionEnum.RES_720P


def test_override_with_multiple_dots_relative_path():
    from tests.support_package.b import B
    from tests.support_package.subpkg.a import A

    env_cfg = cfn.Config(Env, camera=A)

    env_obj = env_cfg.override(camera='...b.B').instantiate()

    assert env_obj.camera is B


def test_override_with_dot_from_cfg_module_applies_relative_to_cfg_module():
    from tests.support_package.cfg import a_cfg_value1

    env_cfg = cfn.Config(Env, camera=a_cfg_value1)

    env_obj = env_cfg.override(camera='.a_cfg_value2').instantiate()

    assert env_obj.camera.value == 2


def test_override_with_dot_from_copied_config_applies_relative_to_cfg_module():
    from tests.support_package.cfg2 import a_cfg_value1_copy

    env_cfg = cfn.Config(Env, camera=a_cfg_value1_copy)

    env_obj = env_cfg.override(camera='.return2').instantiate()

    assert env_obj.camera == 2


def test_override_with_dot_decorator_config_applies_relative_to_cfg_module():
    from tests.support_package.cfg2 import return1

    env_cfg = cfn.Config(Env, camera=return1)

    env_obj = env_cfg.override(camera='.return2').instantiate()

    assert env_obj.camera == 2


def test_config_callable_with_dot_from_copied_config_applies_relative_to_cfg_module():
    from tests.support_package.cfg2 import a_cfg_value1_copy

    env_cfg = cfn.Config(Env, camera=a_cfg_value1_copy)

    env_obj = env_cfg(camera='.return2')

    assert env_obj.camera == 2


def test_override_nesetd_value_with_dot_from_copied_config_applies_relative_to_cfg_module():
    from tests.support_package.cfg2 import a_nested_b_value1

    env_cfg = cfn.Config(Env, camera=a_nested_b_value1)

    env_obj = env_cfg.override(**{'camera.value': '.b_cfg_value2'}).instantiate()

    assert env_obj.camera.value.value == 2


def test_override_with_dot_from_overriden_config_applies_relative_to_cfg_module():
    from tests.support_package.cfg2 import a_cfg_value1_override_value3

    env_cfg = cfn.Config(Env, camera=a_cfg_value1_override_value3)

    env_obj = env_cfg.override(camera='.return2').instantiate()

    assert env_obj.camera == 2


def test_override_with_dot_and_string_default():
    env_cfg = cfn.Config(Env).override(camera='@tests.test_config.Camera')
    env_obj = env_cfg.override(camera='.static_object').instantiate()

    assert env_obj.camera is static_object


def test_override_with_dot_without_default_treated_as_literal():
    env_cfg = cfn.Config(Env, camera=None)

    # Leading-dot strings are treated as literals when default is not a Config
    cfg2 = env_cfg.override(camera='.static_object')
    obj = cfg2.instantiate()
    assert obj.camera == '.static_object'


def test_override_import_error_wrapped_with_context():
    env_cfg = cfn.Config(Env, camera=cfn.Config(Camera, name='OpenCV'))

    with pytest.raises(cfn.ConfigError) as exc_info:
        env_cfg.override(camera='@not.a.real.path')

    assert str(exc_info.value) == "Failed to override 'camera' with value '@not.a.real.path'"
    assert isinstance(exc_info.value.__cause__, ImportError)
    assert 'not.a.real.path' in str(exc_info.value.__cause__)


def test_override_nested_error_preserves_original_cause():
    env_cfg = cfn.Config(Env, camera=cfn.Config(Camera, name='OpenCV'))

    with pytest.raises(cfn.ConfigError) as exc_info:
        env_cfg.override(**{'camera.name.nonexistent': 'value'})

    assert str(exc_info.value) == "Failed to override 'name' with value 'value'"
    assert isinstance(exc_info.value.__cause__, cfn.ConfigError)
    assert 'Cannot set value' in str(exc_info.value.__cause__)


def test_relative_import_with_enum_default():
    """Test that relative imports work when the default is an Enum value."""
    import http

    def process_enum(enum_val):
        return enum_val.value

    # Set up a config with an enum default
    cfg = cfn.Config(process_enum, enum_val=http.HTTPStatus.OK)

    # Override with a relative import (this should work after the fix)
    result_cfg = cfg.override(enum_val='.NOT_FOUND')
    result = result_cfg.instantiate()

    assert result == 404


def test_relative_import_with_nested_enum_default():
    """Test that relative imports work when the default is a nested Enum value.

    Example: geom.Rotation.Representation.ROTVEC
    """
    import http

    def process_nested_enum(enum_val):
        return enum_val.name

    # Set up a config with a nested enum default (HTTPStatus is nested in http module)
    cfg = cfn.Config(process_nested_enum, enum_val=http.HTTPStatus.OK)

    # Override with a relative import (this should work after the fix)
    result_cfg = cfg.override(enum_val='.BAD_REQUEST')
    result = result_cfg.instantiate()

    assert result == 'BAD_REQUEST'


def identity(x):
    return x


def test_override_with_dot_resolves_against_nested_config_module():
    """
    After we copy a Config that originated in cfg2.py, overriding it
    with a relative import ('.return2') must still resolve inside the
    *cfg2* module, not configuronic.config.
    """
    top_cfg = cfn.Config(identity, x=tests.support_package.cfg2.a_cfg_value1_copy)

    result = top_cfg.override(x='.return2').instantiate()
    assert result == 2


def test_config_callable_with_dot_resolves_against_nested_config_module():
    top_cfg = cfn.Config(identity, x=tests.support_package.cfg2.a_cfg_value1_copy)
    result = top_cfg(x='.return2')
    assert result == 2


def test_config_objects_are_callable():
    """Test that Config objects and overridden Config objects are callable."""

    @cfn.config(a=1, b=2)
    def add_numbers(a, b):
        return a + b

    # Base config should be callable
    assert callable(add_numbers)
    assert add_numbers() == 3

    # Overridden config should also be callable
    overridden_config = add_numbers.override(a=5)
    assert callable(overridden_config)
    assert overridden_config() == 7

    # Calling with additional override should work
    assert add_numbers(a=10, b=20) == 30

    # Calling overridden config with additional override should work
    assert overridden_config(b=10) == 15


def test_string_override_with_leading_dot():
    @cfn.config(greeting='Hello', name='world')
    def greeting(greeting, name):
        return f'{greeting}, {name}!'

    # Because the default is a string, the leading dot is not treated as a relative import
    assert greeting.override(name='.weird_name')() == 'Hello, .weird_name!'


def test_config_with_list_arg_could_be_overridden():
    @cfn.config(a=['a', 'b', 'c'])
    def join(a):
        return ''.join(a)

    assert join.override(**{'a.0': 'X'}).instantiate() == 'Xbc'


def test_config_with_list_as_arg_preserve_list_type():
    @cfn.config(a=[1, 2, 3])
    def return_arg(a):
        return a

    assert isinstance(return_arg.override(**{'a.0': 100}).instantiate(), list)


def test_config_with_list_arg_with_nested_config_could_be_overridden():
    @cfn.config(value=1)
    def obj1(value):
        return value

    @cfn.config(value=2)
    def obj2(value):
        return value

    @cfn.config(arg=[obj1, obj2])
    def return_list(arg):
        return arg

    assert return_list.override(**{'arg.0.value': 100}).instantiate() == [100, 2]


def test_override_list_entry_resolves_relative_dot():
    import tests.support_package.cfg2 as cfg2

    @cfn.config(arg=[cfg2.a_cfg_value1_copy])
    def return_first(arg):
        return arg[0]

    # Override nested list index with a relative path; should resolve to cfg2.return2
    assert return_first.override(**{'arg.0': '.return2'}).instantiate() == 2


def test_override_list_entry_resolves_absolute_at():
    import tests.support_package.cfg2 as cfg2

    @cfn.config(arg=[cfg2.a_cfg_value1_copy])
    def return_first(arg):
        return arg[0]

    # Override nested list index with absolute import path
    assert return_first.override(**{'arg.0': '@tests.support_package.cfg2.return2'}).instantiate() == 2


def test_config_with_tuple_arg_with_nested_config_could_be_overridden():
    @cfn.config(value=1)
    def obj1(value):
        return value

    @cfn.config(value=2)
    def obj2(value):
        return value

    @cfn.config(arg=(obj1, obj2))
    def return_tuple(arg):
        return arg

    assert return_tuple.override(**{'arg.0.value': 100}).instantiate() == (100, 2)


def test_config_with_dict_arg_could_be_overridden():
    @cfn.config(arg={'a': 1, 'b': 2})
    def return_dict(arg):
        return arg

    assert return_dict.override(**{'arg.a': 100}).instantiate() == {'a': 100, 'b': 2}


def test_config_with_dict_arg_with_nested_config_could_be_overridden():
    @cfn.config(value=1)
    def obj1(value):
        return value

    @cfn.config(value=2)
    def obj2(value):
        return value

    @cfn.config(arg={'a': obj1, 'b': obj2})
    def return_dict(arg):
        return arg

    assert return_dict.override(**{'arg.a.value': 100}).instantiate() == {'a': 100, 'b': 2}


def test_override_dict_entry_resolves_relative_dot():
    import tests.support_package.cfg2 as cfg2

    @cfn.config(arg={'x': cfg2.a_cfg_value1_copy})
    def return_x(arg):
        return arg['x']

    # Override nested dict key with a relative path; should resolve to cfg2.return2
    assert return_x.override(**{'arg.x': '.return2'}).instantiate() == 2


def test_override_dict_entry_resolves_absolute_at():
    import tests.support_package.cfg2 as cfg2

    @cfn.config(arg={'x': cfg2.a_cfg_value1_copy})
    def return_x(arg):
        return arg['x']

    # Override nested dict key with absolute import path
    assert return_x.override(**{'arg.x': '@tests.support_package.cfg2.return2'}).instantiate() == 2


def test_required_args_with_no_default_values_returns_all_args():
    @cfn.config()
    def func(a, b):
        return a + b

    assert cfn.get_required_args(func) == ['a', 'b']


def test_required_args_with_default_value_in_function():
    @cfn.config()
    def func(a, b=1):
        return a + b

    assert cfn.get_required_args(func) == ['a']


def test_required_args_with_default_value_in_config():
    @cfn.config(b=1)
    def func(a, b):
        return a + b

    assert cfn.get_required_args(func) == ['a']


def test_required_args_with_default_value_in_config_and_function_returns_all_args():
    @cfn.config(a=1)
    def func(a, b=1):
        return a + b

    assert cfn.get_required_args(func) == []


def test_required_args_with_args_returns_necessary_args():
    def func(a, b, c):
        return a + b + c

    func_cfg = cfn.Config(func, 1, 2)

    assert cfn.get_required_args(func_cfg) == ['c']


def test_required_args_with_args_and_keyword_only_args_returns_necessary_args():
    def func(a, b, *, c):
        return a + b + c

    func_cfg = cfn.Config(func, 1, 2)

    assert cfn.get_required_args(func_cfg) == ['c']


def test_required_args_with_args_not_required():
    @cfn.config()
    def func(*args_name):
        return sum(args_name)

    assert cfn.get_required_args(func) == []


def test_required_args_with_kwargs_not_required():
    @cfn.config()
    def func(**kwargs_name):
        return sum(kwargs_name.values())

    assert cfn.get_required_args(func) == []


def test_required_args_with_nested_config():
    @cfn.config()
    def nested_func(req_arg):
        pass

    @cfn.config(a=nested_func)
    def func(a):
        pass

    assert cfn.get_required_args(func) == ['a.req_arg']


def test_required_args_with_nested_config_in_list():
    @cfn.config()
    def nested_func(req_arg):
        pass

    @cfn.config(a=1)
    def complete_config(a):
        pass

    @cfn.config(a=[complete_config, nested_func])
    def func(a):
        pass

    assert cfn.get_required_args(func) == ['a.1.req_arg']


def test_required_args_with_nested_config_in_dict():
    @cfn.config()
    def nested_func(req_arg):
        pass

    @cfn.config(a=1)
    def complete_config(a):
        pass

    @cfn.config(arg={'a': complete_config, 'b': nested_func})
    def func(arg):
        pass

    assert cfn.get_required_args(func) == ['arg.b.req_arg']


def test_override_existing_list_arg_reports_context_for_index_error():
    @cfn.config(a=[1, 2, 3])
    def func(a):
        return a

    with pytest.raises(cfn.ConfigError) as exc_info:
        func.override(**{'a.4': 4})

    assert str(exc_info.value) == "Failed to override 'a' with value '4'"
    assert isinstance(exc_info.value.__cause__, IndexError)
    assert 'list assignment index out of range' in str(exc_info.value.__cause__)


def test_override_non_existing_list_arg_raises_config_error():
    @cfn.config(a=[1, 2, 3])
    def func(a):
        return a

    with pytest.raises(cfn.ConfigError) as exc_info:
        func.override(**{'b.0': 4})

    assert str(exc_info.value) == "Failed to override 'b' with value '4'"
    assert isinstance(exc_info.value.__cause__, cfn.ConfigError)
    assert "Argument 'b' not found in config" in str(exc_info.value.__cause__)


def test_override_kwargs_function_with_new_argument_returns_expected_dict():
    @cfn.config()
    def func(**kwargs):
        return kwargs

    res = func.override(a=4).instantiate()

    assert res == {'a': 4}


def test_override_non_existing_nested_argument_raises_config_error():
    @cfn.config(a=1)
    def obj1(a):
        return a

    @cfn.config(b=obj1)
    def obj2(b):
        return b

    @cfn.config(arg=obj2)
    def composite_obj(arg):
        return arg

    with pytest.raises(cfn.ConfigError) as exc_info:
        composite_obj.override(**{'arg.a.b': 4}).instantiate()

    assert str(exc_info.value) == "Failed to override 'a' with value '4'"
    assert isinstance(exc_info.value.__cause__, cfn.ConfigError)
    assert "Argument 'arg.a' not found in config" in str(exc_info.value.__cause__)


def test_escape_at_sign_with_double_at():
    """Test that @@ escapes to literal @ in string values."""

    @cfn.config(message='Hello')
    def print_message(message):
        return message

    # Test that @@ becomes literal @
    result = print_message.override(message='@@this_is_literal_at_sign').instantiate()
    assert result == '@this_is_literal_at_sign'


def test_escape_at_sign_in_nested_config():
    """Test that @@ escaping works in nested configurations."""

    @cfn.config(name='default')
    def create_person(name):
        return {'name': name}

    @cfn.config(person=create_person)
    def create_profile(person):
        return person

    result = create_profile.override(**{'person.name': '@@special_name'}).instantiate()
    assert result['name'] == '@special_name'


def test_escape_at_sign_with_multiple_at_signs():
    """Test that multiple @ signs are handled correctly."""

    @cfn.config(text='default')
    def process_text(text):
        return text

    # Test multiple @ signs
    result = process_text.override(text='@@@three_at_signs').instantiate()
    assert result == '@@three_at_signs'


def test_config_args_kwargs_overlap_produces_error():
    def func(a, b):
        return a + b

    func = cfn.Config(func, 1, 2, b=3)

    with pytest.raises(TypeError, match=".* got multiple values for argument 'b'.*"):
        func()


def test_config_instantiation_string_in_decorator_resolves_properly():
    @cfn.config(a='@tests.support_package.cfg2.return1')
    def func(a):
        return a

    assert func.instantiate() == 1


def test_config_instantiation_string_in_class_resolves_properly():
    def func(a):
        return a

    func_cfg = cfn.Config(func, a='@tests.support_package.cfg2.return1')

    assert func_cfg.instantiate() == 1


def test_config_instantiation_string_positional_arg_in_class_resolves_properly():
    def func(a):
        return a

    func_cfg = cfn.Config(func, '@tests.support_package.cfg2.return1')

    assert func_cfg.instantiate() == 1


def test_config_copy_with_double_at_sign_in_decorator_produces_same_config():
    @cfn.config(a='@@some_string')
    def func(a):
        return a

    func_copy = func.copy()

    assert func_copy() == func() == '@some_string'


def test_config_copy_with_double_at_sign_in_class_positional_arg_produces_same_config():
    def func(a):
        return a

    func_cfg = cfn.Config(func, '@@some_string')

    func_copy = func_cfg.copy()

    assert func_copy() == func_cfg() == '@some_string'


def test_config_dot_string_in_decorator_is_literal():
    @cfn.config(a='.return1')
    def func(a):
        return a

    assert func() == '.return1'


def test_config_dot_string_in_class_is_literal():
    def func(x):
        return x

    cfg = cfn.Config(func, x='.return1')
    assert cfg() == '.return1'


def test_config_dot_string_positional_arg_in_class_is_literal():
    def func(x):
        return x

    cfg = cfn.Config(func, '.return1')
    assert cfg() == '.return1'


# ========================================
# Tests for array and dict config references
# ========================================


def test_override_with_list_of_absolute_references():
    """Test passing a list with absolute (@) import paths via override."""

    @cfn.config()
    def return_items(items):
        return items

    items = ['@tests.support_package.cfg2.return1', '@tests.support_package.cfg2.return2']
    result = return_items.override(items=items)
    result = result.instantiate()

    # Configs are resolved to Config objects, then instantiated to their return values
    assert result == [1, 2]


def test_override_with_list_of_mixed_references():
    """Test list override with mixed absolute and relative references.

    When overriding entire list, all relative paths resolve to config's creator module,
    NOT to individual element defaults. This is true even when defaults exist.
    """
    import tests.support_package.cfg3 as cfg3

    # cfg3.process_items has default from cfg2, but when overriding entire list,
    # all relative paths resolve to cfg3 (where config is defined)
    result = cfg3.process_items.override(
        items=['.func_a', '@tests.support_package.cfg2.return1', '.func_b', '.func_c']
    ).instantiate()

    # All relative paths resolve to cfg3, not to element defaults
    assert result == ['a', 1, 'b', 'c']


def test_override_with_empty_list():
    """Test overriding with an empty list."""

    @cfn.config(items=[1, 2, 3])
    def return_items(items):
        return items

    result = return_items.override(items=[]).instantiate()
    assert result == []


def test_override_with_list_of_configs():
    """Test passing a list of Config objects directly."""
    import tests.support_package.cfg as cfg

    @cfn.config()
    def return_items(items):
        return items

    result = return_items.override(items=[cfg.a_cfg_value1, cfg.b_cfg_value1]).instantiate()

    # Configs should be instantiated
    assert len(result) == 2
    assert result[0].value == 1
    assert result[1].value == 1


def test_override_with_dict_of_absolute_references():
    """Test passing a dict with absolute (@) import paths via override."""

    @cfn.config()
    def return_config(config):
        return config

    result = return_config.override(
        config={'func1': '@tests.support_package.cfg2.return1', 'func2': '@tests.support_package.cfg2.return2'}
    ).instantiate()

    assert result == {'func1': 1, 'func2': 2}


def test_override_with_dict_of_relative_references():
    """All items in dict override resolve relative to config's creator module.

    cfg3.process_config is defined in cfg3 but has default from cfg2.
    When overriding the entire dict, ALL relative paths resolve to cfg3.
    """
    import tests.support_package.cfg3 as cfg3

    result = cfg3.process_config.override(
        config={'key1': '@tests.support_package.cfg2.return1', 'key2': '.func_a', 'key3': '.func_c'}
    ).instantiate()

    # All relative paths resolve to cfg3 (config's creator module), not to key defaults
    assert result == {'key1': 1, 'key2': 'a', 'key3': 'c'}


def test_override_with_nested_collections():
    """Test nested lists and dicts with config references."""

    @cfn.config()
    def return_items(items):
        return items

    result = return_items.override(
        items=[
            ['@tests.support_package.cfg2.return1', '@tests.support_package.cfg2.return2'],
            {'a': '@tests.support_package.cfg2.return1', 'b': 42},
        ]
    ).instantiate()

    assert result[0] == [1, 2]
    assert result[1] == {'a': 1, 'b': 42}


def test_list_override_then_indexed_override():
    """Test that list override works, then indexed override can modify it."""
    import tests.support_package.cfg as cfg

    @cfn.config()
    def return_items(items):
        return items

    # First override with list, then indexed override can modify it
    cfg_final = return_items.override(**{
        'items': ['@tests.support_package.cfg2.return1', '@tests.support_package.cfg2.return2'],
        'items.0': cfg.a_cfg_value1,
    })

    result = cfg_final.instantiate()
    assert len(result) == 2
    assert isinstance(result[0], tests.support_package.subpkg.a.A)
    assert result[0].value == 1
    assert result[1] == 2


def test_list_override_with_literals():
    """Test that literal values (strings, numbers, etc.) work in lists."""

    @cfn.config()
    def return_items(items):
        return items

    result = return_items.override(items=['hello', 42, 3.14, None]).instantiate()
    assert result == ['hello', 42, 3.14, None]


def test_dict_override_with_mixed_types():
    """Test dict with mixed value types."""

    @cfn.config()
    def return_config(config):
        return config

    result = return_config.override(
        config={'string': 'hello', 'number': 42, 'ref': '@tests.support_package.cfg2.return1', 'none': None}
    ).instantiate()

    assert result['string'] == 'hello'
    assert result['number'] == 42
    assert result['ref'] == 1
    assert result['none'] is None


def test_list_with_escaped_at_prefix():
    """Test that @@ escapes to literal @ in lists."""

    @cfn.config()
    def return_items(items):
        return items

    result = return_items.override(items=['@@not.a.reference', 'normal']).instantiate()
    assert result == ['@not.a.reference', 'normal']


def test_literal_dot_prefix_strings_via_indexed_override():
    """Test passing literal dot-prefixed strings by using indexed override workaround.

    Full list override with '.' strings attempts relative resolution.
    To pass literals like './data' or '.env', use indexed override or
    initialize with empty strings and override later.
    """

    @cfn.config(items=['', ''])
    def return_items(items):
        return items

    # Override individual indices with literal dot-prefixed strings
    result = return_items.override(**{'items.0': './data', 'items.1': '.env'}).instantiate()
    assert result == ['./data', '.env']


def test_nested_references_resolved_at_all_depths():
    """Test that config references are resolved at all nesting levels.

    Both absolute (@) and relative (.) references should be resolved
    recursively throughout nested collections.
    """
    # Define in cfg3 module context so relative imports resolve there
    import tests.support_package.cfg3 as cfg3

    # Both top-level and nested references should resolve
    result = cfg3.return_items.override(
        items=[
            '.func_a',  # Top-level relative: resolves to cfg3.func_a
            ['@tests.support_package.cfg2.return1', '.func_b'],  # Nested: both resolve
            {'abs': '@tests.support_package.cfg2.return2', 'rel': '.func_c'},  # Nested dict: both resolve
        ]
    ).instantiate()

    assert result[0] == 'a'  # Top-level relative resolved
    assert result[1] == [1, 'b']  # Nested absolute and relative resolved
    assert result[2] == {'abs': 2, 'rel': 'c'}  # Nested dict references resolved


def test_relative_import_uses_spec_name_when_module_name_is_main():
    result = subprocess.run(
        [sys.executable, '-m', 'tests.support_package.main_runner'],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0, f'stdout: {result.stdout}\nstderr: {result.stderr}'


if __name__ == '__main__':
    pytest.main()



import sys

from configuronic.cli import cli
from configuronic.config import Config, _import_object_from_path


def main():
    if len(sys.argv) < 2 or not sys.argv[1].startswith('@'):
        print('Usage: python -m configuronic @path.to.module.Config [--param=value ...]', file=sys.stderr)
        sys.exit(1)

    import_path = sys.argv[1]
    sys.argv = [sys.argv[0]] + sys.argv[2:]

    obj = _import_object_from_path(import_path)

    if not isinstance(obj, Config):
        print(f'Error: {import_path} resolved to {type(obj).__name__}, expected a Config object', file=sys.stderr)
        sys.exit(1)

    cli(obj)


if __name__ == '__main__':
    main()



import subprocess
import sys


def run_main(*args):
    return subprocess.run([sys.executable, '-m', 'configuronic', *args], capture_output=True, text=True)


def test_main_resolves_and_runs_config():
    result = run_main('@tests.support_package.cfg.echo')
    assert result.returncode == 0
    assert result.stdout.strip() == 'hello'


def test_main_passes_override_params():
    result = run_main('@tests.support_package.cfg.echo', '--message=world')
    assert result.returncode == 0
    assert result.stdout.strip() == 'world'


def test_main_missing_arg_prints_usage():
    result = run_main()
    assert result.returncode == 1
    assert 'Usage:' in result.stderr


def test_main_arg_without_at_prefix_prints_usage():
    result = run_main('tests.support_package.cfg.echo')
    assert result.returncode == 1
    assert 'Usage:' in result.stderr


def test_main_invalid_import_path():
    result = run_main('@nonexistent.module.thing')
    assert result.returncode != 0


def test_main_non_config_object():
    result = run_main('@tests.support_package.b.B')
    assert result.returncode == 1
    assert 'expected a Config object' in result.stderr



from unittest.mock import patch

import pytest

import configuronic as cfn


def test_cli_kwarg_override_overrides_value(capfd):
    @cfn.config(a=1)
    def identity(a):
        print(a)

    with patch('sys.argv', ['script.py', '--a=2']):
        cfn.cli(identity)
        out, err = capfd.readouterr()
        assert out == '2\n'


def test_cli_help_prints_has_required_args(capfd):
    @cfn.config()
    def identity(a, b):
        print(a, b)

    with patch('sys.argv', ['script.py', '--help']):
        cfn.cli(identity)
        out, err = capfd.readouterr()
        assert "a: <REQUIRED>" in out
        assert "b: <REQUIRED>" in out


def test_cli_help_prints_docstring(capfd):
    @cfn.config()
    def identity(a, b):
        """This is a test function.
        """
        print(a, b)

    with patch('sys.argv', ['script.py', '--help']):
        cfn.cli(identity)
        out, err = capfd.readouterr()
        assert "This is a test function." in out


def test_cli_help_prints_nested_required_args(capfd):
    @cfn.config()
    def nested_func(req_arg):
        pass

    @cfn.config(a=nested_func)
    def func(a):
        pass

    with patch('sys.argv', ['script.py', '--help']):
        cfn.cli(func)
        out, err = capfd.readouterr()
        assert "a.req_arg: <REQUIRED>" in out


def test_cli_multiple_commands_call_overrides_arg(capfd):
    @cfn.config()
    def func1(a):
        print(f"a: {a}")

    @cfn.config()
    def func2(b):
        print(f"b: {b}")

    with patch('sys.argv', ['script.py', 'func1', '--a=1']):
        cfn.cli({'func1': func1, 'func2': func2})
        out, err = capfd.readouterr()
        assert out == 'a: 1\n'

def test_cli_multiple_commands_help_prints_commands_list(capfd):
    @cfn.config()
    def func1(a):
        """Docstring for func1"""
        print(f"a: {a}")

    @cfn.config()
    def func2(b):
        """Docstring for func2"""
        print(f"b: {b}")

    with patch('sys.argv', ['script.py', '--help']):
        cfn.cli({'func1': func1, 'func2': func2})
        out, err = capfd.readouterr()
        assert "python script.py func1 --a=<REQUIRED> # Docstring for func1" in out
        assert "python script.py func2 --b=<REQUIRED> # Docstring for func2" in out


def test_cli_multiple_commands_no_docstring_help_prints_commands_list(capfd):
    @cfn.config()
    def func_no_docstring(a):
        print(f"a: {a}")

    with patch('sys.argv', ['script.py', '--help']):
        cfn.cli({'func_no_docstring': func_no_docstring})
        out, err = capfd.readouterr()
        assert "python script.py func_no_docstring --a=<REQUIRED>" in out


def test_cli_multiple_commands_help_prints_command_help(capfd):
    @cfn.config()
    def func1(a):
        print(f"a: {a}")

    @cfn.config()
    def func2(b):
        print(f"b: {b}")

    with patch('sys.argv', ['script.py', 'func1', '--help']):
        cfn.cli({'func1': func1, 'func2': func2})
        out, err = capfd.readouterr()
        assert "a: <REQUIRED>" in out


def test_cli_multiple_commands_unknown_command_raises_error(capfd):
    @cfn.config()
    def func1(a):
        print(f"a: {a}")

    @cfn.config()
    def func2(b):
        print(f"b: {b}")

    with patch('sys.argv', ['script.py', 'func3']):
        with pytest.raises(ValueError) as e:
            cfn.cli({'func1': func1, 'func2': func2})
        assert "Command 'func3' not found. Available commands: ['func1', 'func2']" in str(e.value)


def test_cli_multiple_commands_unknown_command_with_help_raises_error(capfd):
    @cfn.config()
    def func1(a):
        print(f"a: {a}")

    @cfn.config()
    def func2(b):
        print(f"b: {b}")

    with patch('sys.argv', ['script.py', 'func3', '--help']):
        with pytest.raises(ValueError) as e:
            cfn.cli({'func1': func1, 'func2': func2})
        assert "Command 'func3' not found. Available commands: ['func1', 'func2']" in str(e.value)


def test_cli_multiple_commands_print_help_if_no_args_provided(capfd):
    @cfn.config()
    def func1(a):
        print(f"a: {a}")

    with patch('sys.argv', ['script.py']):
        cfn.cli({'func1': func1})
        out, err = capfd.readouterr()
        assert "python script.py func1 --a=<REQUIRED>" in out


def test_cli_accepts_leading_dot_strings_as_literals(capfd):
    @cfn.config()
    def echo(a):
        print(a)

    # '../data' should be treated as a literal string
    with patch('sys.argv', ['script.py', '--a=../data']):
        cfn.cli(echo)
        out, err = capfd.readouterr()
        assert out.strip() == '../data'

    # './file' should be treated as a literal string
    with patch('sys.argv', ['script.py', '--a=./file']):
        cfn.cli(echo)
        out, err = capfd.readouterr()
        assert out.strip() == './file'

    # '.env' should be treated as a literal string
    with patch('sys.argv', ['script.py', '--a=.env']):
        cfn.cli(echo)
        out, err = capfd.readouterr()
        assert out.strip() == '.env'



import inspect
import sys

import fire

from configuronic.config import Config


def _get_required_args_recursive(config: Config, prefix: str = '') -> list[str]:
    sig = inspect.signature(config.target)
    required_args = []

    for i, (name, param) in enumerate(sig.parameters.items()):
        if param.default != inspect.Parameter.empty:
            continue
        if param.name in config.kwargs:
            if isinstance(config.kwargs[param.name], Config):
                required_args.extend(_get_required_args_recursive(config.kwargs[param.name], f'{prefix}{name}.'))
            if isinstance(config.kwargs[param.name], list | tuple):
                for i, item in enumerate(config.kwargs[param.name]):
                    if isinstance(item, Config):
                        required_args.extend(_get_required_args_recursive(item, f'{prefix}{name}.{i}.'))
            if isinstance(config.kwargs[param.name], dict):
                for k, v in config.kwargs[param.name].items():
                    if isinstance(v, Config):
                        required_args.extend(_get_required_args_recursive(v, f'{prefix}{name}.{k}.'))
            continue
        if i < len(config.args):
            continue
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            # var positional args are not required
            continue
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            # var keyword args are not required
            continue

        required_args.append(f'{prefix}{name}')
    return required_args


def get_required_args(config: Config) -> list[str]:
    """
    Get the list of required arguments to instantiate the target callable.

    Returns:
        List of required argument names (excluding those with default values and those that are already set).
    """
    return _get_required_args_recursive(config, '')


def _cli_single_command(config: Config, is_command_first: bool = False):
    assert 'help' not in config.kwargs, "Config contains 'help' argument. This is reserved for the help flag."

    def _run_and_help(help: bool = False, **kwargs):
        overriden_config = config.override(**kwargs)
        if help:
            if hasattr(overriden_config.target, '__doc__') and overriden_config.target.__doc__:
                print(overriden_config.target.__doc__)
                print('=' * 140)

            print('Config:')
            for arg in get_required_args(overriden_config):
                print(f'{arg}: <REQUIRED>')
            print()
            print(str(overriden_config))
        else:
            return overriden_config.instantiate()

    if is_command_first:
        # HACK: Add fake first argument to a function signature so the `Fire` threat it as a positional argument
        def _run_and_help_wrapper(_command: str=None, help: bool = False, **kwargs):
            return _run_and_help(help, **kwargs)

        return _run_and_help_wrapper
    else:
        return _run_and_help


def _cli_multiple_commands(commands_config: dict[str, Config]):
    def _run_and_help(_command: str = None, help: bool = False, **kwargs):
        args = sys.argv[1:]
        # Handle help flag manually to choose a proper function for fire.Fire
        if len(args) == 0 or args[0] == '--help':
            print('Commands:')
            for command in commands_config:
                target_command_doc = commands_config[command].target.__doc__
                if target_command_doc:
                    first_line_of_doc = target_command_doc.split('\n')[0]
                    command_description = f' # {first_line_of_doc}'
                else:
                    command_description = ''
                required_args = get_required_args(commands_config[command])
                required_args_str = ' '.join([f'--{arg}=<REQUIRED>' for arg in required_args])
                print(f'python {sys.argv[0]} {command} {required_args_str}{command_description}')
            print()
        elif args[0] in commands_config:
            return _cli_single_command(commands_config[args[0]], is_command_first=True)(_command, help, **kwargs)
        else:
            raise ValueError(f"Command '{args[0]}' not found. Available commands: {list(commands_config.keys())}")

    fire.Fire(_run_and_help)


def cli(config: Config | dict[str, Config]):
    """
    Run a config object(s) as a CLI.

    Args:
        config: The config or dict of configs to run. If dict is provided, each key is a command name
         and each value is a config object.

    Example:
        >>> @cfn.config()
        >>> def sum(a, b):
        >>>     return a + b
        >>> cfn.cli(sum)
        >>> # Shell call: python script.py --a 1 --b 2
        >>> # Shell call: python script.py --help

        >>> @cfn.config()
        >>> def sum(a, b):
        >>>     return a + b
        >>> @cfn.config()
        >>> def product(a, b):
        >>>     return a * b
        >>> cfn.cli({'sum': sum, 'product': product})
        >>> # Shell call: python script.py sum --a 1 --b 2
        >>> # Shell call: python script.py product --a 1 --b 2
        >>> # Shell call: python script.py --help
        >>> # Shell call: python script.py sum --help
    """

    if isinstance(config, dict):
        return _cli_multiple_commands(config)
    else:
        fire.Fire(_cli_single_command(config))
